#!/usr/bin/env python3
""" Copyright 2020 Phil Chandler
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""novag_xboard.py v1.2
    This script is designed to act as a pseudo engine in Xboard, so that you can play against a Novag Sapphire, Diamond and later models
    using the Xboard interface.  It has been tested against a Sapphire, Diamond II and Citrine.  I imagine it could be easily modified
    to work with earlier models such as the Super VIP, with suitable work arounds for variations in the Novag serial protocol.
    
    Prerequisites are Python3 and the PySerial library. Since this is written in Python, theoretically, it
    should work under Windows, but that hasn't been tested. In Windows the built in support for Novag computers in Arena is recommended,
    and the UCI engines created by Alain Zanchetta: http://alain.zanchetta.free.fr/softs/Alainza.Uci.1.0.335.1.Setup.msi
    
    For a Sapphire or Diamond you will need a suitable hardware adapter to connect to a
    PC serial port, or one of the original Novag Super System Distributor hubs.  You can buy a modern adapter from https://sites.google.com/site/bergersprojects/chess_accessories/novag-arena_interface or build
    one yourself following instructions from https://www.schach-computer.info/wiki/index.php?title=Novag_Distributor_Replacement.
    
    This script was developed on a PC running Linux (Ubuntu 20.04 as it happens). Different generations of machines have 
    slight differences in their serial protocol, so some adjustments are needed. The Sapphire doesn't, for example, support
    the 'J' command listed in the documentation to tell the Novag to 'Go'. You need to use the actual Go button to play black, or swap
    colours mid-game.  Essentially you tell Xboard the engine is now playing the other colour and then press Go on the Novag.

    The script takes the serial port as a command line argument, and then works out what baud rate to use for the computer connected.
    If you want to play against a computer you ass it as an 1st engine in Xboard, giving the command for the engine as something like:

    "<path to script>/novag_xboard.py /ttyUSB0"

    The script will get the model name from the computer and tell Xboard to use that as the machine's nickname.

    For computer vs computer games, (assuming you have a second serial port available), a good way of starting Xboard is with a command
    something like this:

    xboard -fcp "<path to script>/novag_xboard.py /dev/ttyUSB0" -fd "/<path to script>" -scp "<path to script>/novag_xboard.py /dev/ttyUSB1" -sd "<path to script>"

    -fcp/-fd define the parameters for the 1st engine and -scp/-sd for the second.  To change the colours, swap the serial ports.

    Start the game by selecting 'Two Machines' from the Xboard Mode menu.
    
    Any bug reports, improvements or feedback, please send to user kalatov at https://hiarcs.net/forums/ For outputing detailed debugging
    information see the notes below"""

import serial, time, sys
# Set debug = True to send additional output to Xboard for logging in debug mode
# debug2 = True gives more indepth information
# To see debug info run Xboard with the command 'xboard -debug -engineDebugOutput 2'
# and tail -f the xboard.debug file in the directory you are running the script from
debug = True
debug2 = True
name = ""
colour = "black" #default unless changed later
port = str(sys.argv[1])
if debug2:
    print(port)
baudrate = 9600 # set as default, the script will check if this works later

def send_instruction(instruction, **kwargs):
    if debug2:
        print(ser)
    compresponse = ""
    ser.write((instruction + '\r\n').encode('utf-8')) # Send the instruction
    while ser.in_waiting == 0:
        waiting = True
    while ser.in_waiting > 0:
        response = ser.read_until() # Get the computer's response
        responsestr =str(response)
        if debug:
            print("raw response from computer: " + responsestr)
        checkendcode = check_endcode(responsestr)
        if checkendcode != "nocode":
            compresponse = checkendcode
        elif instruction[0] == 'M': #It's a move, consume echo and get computer's move
            if name == "UCBII": #Computer is a Citrine
                ser.write((instruction + '\r\n').encode('utf-8')) # Send the instruction a 2nd time otherwise Citrine waits for piece to be moved.
            response = ser.read_until()
            responsearray = response.split()
            responsestr =str(responsearray)
            if debug:
                print("raw response from computer: " + responsestr)
            if checkendcode != "nocode":
                compresponse = checkendcode
            else:
                #responsearray = responsestr.split()
                if debug2:
                    print("Responsearray:")
                    print(responsearray)
                move1 = str(responsearray[2])
                move2 = move1.split('\'')
                move = move2[1]
                time.sleep(1.0)
                compresponse = "move " + move
        elif instruction[0] == 'L': #We're setting the level so need to consume the extra line returned
            response = ser.read_until()
            compresponse =str(response)
        elif instruction[0] == 'J': #Sapphire doesn't recognise J, needs user to press Go
            responsearray = responsestr.split()
            if debug2:
                print("Responsearray:")
                print(responsearray)
            move = responsearray[2]
            compresponse = "move " + move
        elif instruction[0] == 'I':
            idarray = responsestr.split()
            compresponse = idarray[1]
        if debug:
            print("Response from computer: " + compresponse)
        print(compresponse)
    return compresponse

def set_option(xbinput):
    optionarray = xbinputarray[1].split("=")
    option = optionarray[0]
    value = optionarray[1]
    if option == "Level":
        response = send_instruction("L "+ value)
        optionset = "Level set to " + response[10 : 14]
    # Add other options as elseifs, ending by setting optionset
    # Options are defined in the 'feature' line sent in response to 'protover' from Xboard.
    return optionset

def send_message(message):
    print("telluser " + message)
    return

def check_endcode(responsestr):
    checkendcode = responsestr.find("M#")
    if checkendcode != -1: # It's a code for how the game ended
        message = responsestr[checkendcode : checkendcode + 3]
        if debug:
            print("message code: " + message)
        if message == 'M#1':
            result = "1/2-1/2 {Draw by repetition}"
        elif message == 'M#2':
            result = "1/2-1/2 {Draw by 50 move rule}"
        elif message == 'M#3':
            result = "1/2-1/2 {Draw by insufficient material}"
        elif message == 'M#4':
            result = "1/2-1/2 {Draw by stalemate}"
        elif message == 'M#5':
            if colour == "white":
                result = "1-0 {white mates}"
            elif colour == "black":
                result = "0-1 {black mates}"
        elif message == 'M#6':
            if colour == "white":
                result = "0-1 {white resigns}"
            elif colour == "black":
                result = "1-0 {black resigns}"
    else:
        result = "nocode"
    print("check end code: " + result)
    return result

ser = serial.Serial(port, baudrate, timeout=3.0)  # open serial port
#Test if we have communications
if debug2:
    if ser.is_open:
        print("Port is open")
# Switch Xmit on and get the device name
ser.write(("X on" + '\r\n').encode('utf-8'))
response = ser.read_until() # Get the computer's response
commstest =str(response)
if commstest != "b'Xmit on\\r\\n'": #9600 didn't work, try 57600
    ser.close()
    ser.baudrate = 57600
    ser.open()
    if debug2:
        print(ser)
        if ser.is_open:
            print("Port is open")
    # This is probably a Citrine that takes a lot to wake it up
    ser.write(("X on" + '\r\n').encode('utf-8'))
    ser.read_until()
    ser.write(("X on" + '\r\n').encode('utf-8'))
    ser.read_until()
    ser.write(("X on" + '\r\n').encode('utf-8'))
    response = ser.read_until() # Get the computer's response
    commstest =str(response)
    if commstest != "b'Xmit on\\r\\n'":
        print("No computer detected")
        ser.reset_input_buffer()
        ser.reset_output_buffer()
        ser.close()
        exit()
name = send_instruction("I")
ser.timeout = None #Timeout can cause missed moves 

# Wait for input from Xboard
xbinput = ""
while 1 :
    xbinput = input()
    xbinputarray = xbinput.split()
    command = xbinputarray[0]
    if debug == True:
        print("Command from Xboard was " + xbinput)
    if command == 'quit':
        send_instruction("new")
        ser.reset_input_buffer()
        ser.reset_output_buffer()
        ser.close()
        exit()
    elif command == 'xboard':
        print("\n")
    elif command == 'protover':
        # Set which features accepted and define any engine options configurable in the engine options dialogue of the GUI.
        # The option defined here is the ability to set the computers level, useing the EA1, TR1 etc. notation.
        print("feature usermove=1 ping=0 time=0 draw=0 sigint=0 sigterm=0 option=\"Level -string e.g. TR1\" myname=\""  + name +"\" colors=1 reuse=0")
        print("feature done=1")
    elif command == 'new':
        send_instruction(command)
        print("new")
    elif command == 'black':
        colour = "black"
    elif command == 'white':
        colour = "white"
    elif command == 'go':
        if colour != 'black':
            print(send_instruction("J"))
        # J 'Go' command is not recognised by the Sapphire, so the user needs to press the Go button when changing colours in Xboard
        # In the GUI change which colour the engine is playing and then press the Go button.
    elif command == 'usermove':
        instruction = "M" + xbinputarray[1]
        send_instruction(instruction, colour=colour)
    elif command == 'remove':
        send_instruction("T")
        send_instruction("T")
    elif command == 'option':
        set_option(xbinput)
    elif command == 'I':
        send_instruction("I")
    elif command == 'accepted' or command == 'rejected':
        True #No response
    elif command == 'result':
        # game over, let's flush the serial buffers
        ser.reset_input_buffer()
        ser.reset_output_buffer()
    else:
        print("Error (unknown command): " + xbinput)